"""
    This file is part of Accountant.

    Accountant is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Accountant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with Accountant.  If not, see <http://www.gnu.org/licenses/>.
"""
import dateutil.parser

from flask_restplus import Namespace, Resource, fields, marshal_with_field

from ..models import db
from ..models.accounts import Account
from ..models.operations import Operation

from .users import requires_auth

from ..fields import Object


# pylint: disable=invalid-name
ns = Namespace('operation', description='Operation management')

# Operation with sold model.
operation_model = ns.model('Operation', {
    'id': fields.Integer(
        default=None,
        readonly=True,
        description='Id of the operation'),
    'operation_date': fields.DateTime(
        dt_format='iso8601',
        required=True,
        description='Date of the operation'),
    'label': fields.String(
        required=True,
        description='Label of the operation'),
    'value': fields.Float(
        required=True,
        description='Value of the operation'),
    'pointed': fields.Boolean(
        required=True,
        description='Pointed status of the operation'),
    'category': fields.String(
        required=False,
        default=None,
        description='Category of the operation'),
    'account_id': fields.Integer(
        required=True,
        readonly=True,
        description='Account id of the operation'),
    'scheduled_operation_id': fields.Integer(
        default=None,
        readonly=True,
        description='Scheduled operation ID of the operation'),
    'confirmed': fields.Boolean(
        description='Confirmed status of the operation'),
    'canceled': fields.Boolean(
        description='Canceled status of the operation (for a scheduled one)')
})

operation_with_sold_model = ns.clone(
    'OperationWithSold', operation_model, {
        'sold': fields.Float(
            readonly=True,
            description='Cumulated sold'
        ),
    }
)

# Parser for a date range and an account id.
account_range_parser = ns.parser()
account_range_parser.add_argument(
    'begin',
    type=lambda a: dateutil.parser.parse(a) if a else None,
    required=False,
    default=None,
    location='args',
    help='Begin date of the time period'
)
account_range_parser.add_argument(
    'end',
    type=lambda a: dateutil.parser.parse(a) if a else None,
    required=False,
    default=None,
    location='args',
    help='End date of the time period'
)
account_range_parser.add_argument(
    'account_id',
    type=int,
    required=True,
    location='args',
    help='Id of the account'
)


@ns.route('/')
@ns.doc(
    security='apikey',
    responses={
        401: 'Unauthorized'
    })
class OperationListResource(Resource):
    @requires_auth
    @ns.response(200, 'OK', [operation_with_sold_model])
    @ns.expect(parser=account_range_parser)
    @marshal_with_field(fields.List(Object(operation_with_sold_model)))
    def get(self):
        """
        Get operations with solds for a specific account.
        """
        data = account_range_parser.parse_args()

        return Operation.query(
            begin=data['begin'],
            end=data['end']
        ).filter(
            Operation.account_id == data['account_id']
        ).all(), 200

    @requires_auth
    @ns.response(201, 'Operation created', operation_model)
    @ns.response(404, 'Account not found')
    @ns.response(406, 'Invalid operation data')
    @marshal_with_field(Object(operation_model))
    def post(self):
        """
        Create a new operation.
        """
        data = ns.apis[0].payload

        account_id = data['account_id']
        account = Account.query().get(account_id)

        if not account:
            ns.abort(
                404,
                error_message='Account with id %d not found.' % account_id
            )

        # A new operation MUST NOT have an id;
        if 'id' in data and data['id']:
            ns.abort(
                406,
                error_message='Id must not be provided on creation.'
            )

        operation = Operation(**data)

        db.session.add(operation)

        return operation, 201


@ns.route('/<int:id>')
@ns.doc(
    security='apikey',
    params={
        'id': 'Id of the operation to manage'
    },
    responses={
        401: 'Unauthorized',
        404: 'Operation not found'
    })
class OperationResource(Resource):
    @requires_auth
    @ns.response(200, 'OK', operation_model)
    @marshal_with_field(Object(operation_model))
    def get(self, id):
        """
        Get operation.
        """
        operation = db.session.query(Operation).get(id)

        if not operation:
            ns.abort(
                404,
                error_message='Operation with id %d not found.' % id
            )

        return operation, 200

    @requires_auth
    @ns.expect(operation_model)
    @ns.response(200, 'OK', operation_model)
    @ns.response(406, 'Invalid operation data')
    @marshal_with_field(Object(operation_model))
    def post(self, id):
        data = ns.apis[0].payload

        # Check ID consistency.
        if 'id' in data and data['id'] and data['id'] != id:
            ns.abort(
                406,
                error_message='Id must not be provided or changed on update.'
            )

        operation = db.session.query(Operation).get(id)

        if not operation:
            ns.abort(
                404,
                error_message='Operation with id %d not found.' % id
            )

        # FIXME check account_id consistency.

        # SQLAlchemy objects ignore __dict__.update() with merge.
        for k, v in data.items():
            setattr(operation, k, v)

        db.session.merge(operation)

        return operation, 200

    @requires_auth
    @ns.response(204, 'Operation deleted', operation_model)
    @marshal_with_field(Object(operation_model))
    def delete(self, id):
        operation = db.session.query(Operation).get(id)

        if not operation:
            ns.abort(
                404,
                error_message='Operation with id %d not found.' % id
            )

        db.session.delete(operation)

        return None, 204
